/******************************************************************************

    MODUL
	block.c

    DESCRIPTION
	In this module are all functions that work with blocks.

    NOTES

    BUGS

    TODO

    EXAMPLES

    SEE ALSO

    INDEX

    HISTORY
	27. Aug 1992	ada created

******************************************************************************/

/* Includes */
#include <defs.h>

/* Globale Variable */
Prototype void RedrawBlock	      (void);
Prototype int  is_inblock	      (int, short);
Prototype void displayblock	      (int);
Prototype void text_redrawblock       (int);
Prototype void do_block 	      (void);
Prototype int  block_ok 	      (void);
Prototype void do_bdelete	      (void);
Prototype void do_bcopy 	      (void);
Prototype void do_bmove 	      (void);
Prototype void do_bsource	      (void);

/* Interne Defines & Strukturen */

/* Interne Variable */
static int bs_line, bs_column;
static int cur_line, cur_column;

/* Interne Prototypes */


/* Redraw NORMAL-Block according to cursor-coordinates */
void RedrawBlock (void)
{
    ED * ep = Ep;
    short lines, cols, len;

    /* TODO */
    if (!(BlockFlags & BF_ACTIVE) || (BlockType != BT_NORMAL))
	return;

    if (!(BlockFlags & BF_START) ) {
	/* If nothing happened ... do nothing ! */
	if (BEline == ep->Line && BEcolumn == ep->Column)
	    return;

	/* TODO */
	/* If we move behind out start-position change rules */
	if (ep->Line < BSline || (ep->Line == BSline && ep->Column < BScolumn))
	    return;

	/* Calculate relative offset. Right/Down is positive */
	lines = ep->Line - BEline;
	cols = ep->Column - BEcolumn;

	if (lines > 0) {
	    /* Show all new lines */
	    text_displayseg (BEline - ep->Topline, lines);
	} else if (lines) {
	    /* Erase block in some lines */
	    text_displayseg (ep->Line - ep->Topline, -lines);
	} else {
	    /* we didn't change line but position in the line */
	    RP * rp = ep->Win->RPort;

	    /* Find the real length of our line */
	    len = strlen (ep->List[BEline]);

	    /* ... to the right */
	    if (cols > 0) {
		/* we must advance the block */
		SetAPen (rp, BLOCK_FPEN);
		SetBPen (rp, BLOCK_BPEN);

		/* If we are already beyond EOL, we may stop */
		if (len < BEcolumn)
		    return;

		/* There is a bit of text to display */
		if (len > BEcolumn) {
		    len -= BEcolumn;

		    /* How much (at most) ? */
		    if (len < cols)
			cols = len;

		    /* Display it */
		    Move (rp, ROW(BEcolumn - ep->Topcolumn),
			    COLT(BEline - ep->Topline));
		    Text (rp, Current + BEcolumn, cols);

		    /* Not how much we have already */
		    len -= cols;
		}

		/* If there is length left, user is beyond EOL and we
		   fill up to the right border */
		if (len) {
		    SetAPen (rp, BLOCK_BPEN);

		    RectFill (rp, COL(ep->Column - ep->Topcolumn + 1),
			    ROW(ep->Line), Xpixs, ROW(ep->Line+1)-1);
		}

		/* Reset pens (save) */
		SetAPen (rp, TEXT_FPEN);
		SetBPen (rp, TEXT_BPEN);
	    } else { /* We got shorter */
		SetAPen (rp, TEXT_FPEN);
		SetBPen (rp, TEXT_BPEN);

		len -= ep->Column;

		/* don't do anything if we are still beyong EOL */
		if (len < 0)
		    return;

		/* Make cols positive */
		cols = -cols;

		if (cols < len)
		    len = cols;

		Move (rp, ROW(ep->Column - ep->Topcolumn),
			COLT(BEline - ep->Topline));
		Text (rp, Current + ep->Column, len);

		/* if there is something beyond EOL */
		if (cols > len) {
		    SetAPen (rp, BLOCK_BPEN);

		    RectFill (rp, COL(len + ep->Column + 1),
			    ROW(ep->Line), Xpixs, ROW(ep->Line+1)-1);
		}
	    } /* cols changed */
	} /* lines changed */

	BEline = ep->Line;
	BEcolumn = ep->Column;
    } else {
	/* TODO: What to do if we get above BS */
    }

} /* RedrawBlock */



int is_inblock (int line, short col)
{
    short blockpos = BP_OUTSIDE;

    if (BlockType != BT_NONE && Ep == BEp && !Comlinemode && line >= BSline &&
	    line <= BEline) {

	if (line == BSline)
	    blockpos |= BP_START;
	if (line == BEline)
	    blockpos |= BP_END;

	switch (BlockType) {
	    case BT_LINE:
		blockpos |= BP_INSIDE;
		break;

	    case BT_NORMAL:
		if (col != -1 && (line == BSline && col >= BScolumn) ||
			(line == BEline && col <= BEcolumn))
		    blockpos |= BP_INSIDE;
		break;

	    case BT_VERTICAL:
		if (col != -1 && col >= BScolumn && col <= BEcolumn)
		    blockpos |= BP_INSIDE;
		break;
	} /* switch (BlockType) */
    } /* if validblock */

    return (blockpos);
} /* is_inblock */


void displayblock (int on)
{
    ED	* ep	= Ep;
    short start = ep->Topline;
    short lines = BEline - BSline + 1;
    short flags;

    if (start < BSline)
	start = BSline;

    start -= ep->Topline;

    if (ep == BEp) {
	flags = lineflags (BSline) & lineflags (BEline);

	if (!on && (!flags || flags == LINE_VISIBLE)) {
	    RP * rp = ep->Win->RPort;
	    short	      x;

	    x = lines + start;

	    if (x > Rows)
		x = Rows;

	    SetAPen (rp, TEXT_BPEN);

	    /* Block lschen */
	    if (start < x)
		RectFill (rp, Xbase, ROW(start), Xpixs, ROW(x)-1);

	    /* Block aus */
	    BlockType = BT_NONE;
	}

	/* )ineu darstellen */
	text_displayseg (start, lines);
    } else if (!on) {
	BlockType = BT_NONE;
    }
} /* displayblock */


void text_redrawblock (int ok_IN  WIN * savewin = NULL;

    if (BEp) {
	if (BEp != Ep) {
	    savewin = Ep->Win;
	    text_switch (BEp->Win);
	}

	if (BlockType != BT_NONE) {
	    displayblock (ok_;
	}

	if (savewin)
	    text_switch (savewin);
    } /* if (BEp) */

    if (!ok_ {
	BlockType = BT_NONE;
    }
} /* text_redrawblock */


void do_block()          /* block, unblock   */
{
    text_sync ();

    switch(av[0][0]) {
	case 'b':
	    switch (av[0][5]) { /* block[se] */
		case '\0':
		    if (BlockType == BT_NONE) {
bstart:
			BEp = Ep;
			BSline = Ep->Line;
			BEline = -1;
			BlockType = BT_LINE;	/* Note: DON'T DISPLAY THIS ! */
			title ("Block Begin");
		    } else {
			if (BEline >= 0) {
			    if (Ep->AutoUnblock) {
				text_redrawblock(0);
				goto bstart;
			    } else {
				title("Block Already Marked");
				break;
			    }
			}

			if (BEp != Ep)
			    goto bstart;

			title("Block End");

			BEline = Ep->Line;

			if (BSline > BEline) {
			    BEline = BSline;
			    BSline = Ep->Line;
			}

			BlockType = BT_LINE;	/* Force old-style */
			text_redrawblock (1);   /* Now display */
		    }
		    break;

		case 's': /* blockstart */
		    BlockType = BT_NORMAL;
		    BlockFlags = BF_ACTIVE;
		    BEp = Ep;
		    bs_line = cur_line = BSline = BEline = Ep->Line;
		    bs_column = cur_column = BScolumn = BEcolumn = Ep->Column;

		    /* Block is shown by text_cursor() */
		    break;

		case 'e': /* blockend */
		    if (BlockType == BT_NORMAL)
			BlockFlags = 0; /* Don't resize block anymore */
		    else
			title ("BLOCKEND is for BLOCKSTART");

		    break;
	    }
	    break;

	case 'u':
	    text_redrawblock(0);
	    title ("Block Unmarked");
	    break;

	case 'l':
	    BEp = Ep;
	    BSline = BEline = Ep->Line;
	    title ("Line marked");
	    text_redrawblock (1);
	    break;
    }
} /* do_block() */


int block_ok (void)
{
    if (BlockType != BT_NONE)
	return (1);

    BEp = NULL;
    BSline = BEline = -1;

    title ("Block Not Specified");

    return (0);
} /* block_ok */


void do_bdelete (void)
{
    long n;
    ED *bep = BEp;
    WIN *savewin = Ep->Win;

    /* )ODO */
    if (BlockType != BT_LINE)
	return;

    if (block_ok()) {
	text_switch(bep->Win);

	n = BEline - BSline + 1;

	if (bep->Line >= BSline && bep->Line <= BEline)
	    bep->Line = BSline;

	if (bep->Line > BEline) {
	    bep->Line -= n;
	    bep->Topline -= n;
	}

	freelist(bep->List + BSline, BEline - BSline + 1);
	bmovl(bep->List+BEline+1,bep->List+BSline,(bep->Lines-BEline-1));
	bep->Lines -= n;
	bep->Modified = 1;

	if (bep->Line >= bep->Lines)
	    bep->Line = bep->Lines - 1;

	if (bep->Topline >= bep->Lines)
	    bep->Topline = bep->Lines - 1;

	if (bep->Line < 0)
	    bep->Topline = bep->Line = 0;

	if (bep->Lines == 0) {
	    ++bep->Lines;
	    bep->List[0] = allocline(1);
	}

	text_load ();

	BlockType = BT_NONE;

	if (!text_sync ())
	    text_redisplay ();

	text_switch (savewin);
    }
} /* do_bdelete */


void do_bcopy()
{
    LINE *list;
    long lines, i;
    ED *ep = Ep;
    ubyte * str;

    text_sync ();

    /* TODO */
    if (BlockType != BT_LINE)

    if (!block_ok ())
	return;

    if (ep == BEp && ep->Line > BSline && ep->Line <= BEline) {
	title("Cannot Move into self");
	return;
    }

    lines = BEline - BSline + 1;

    if (extend (ep, lines)) {
	if (list = (LINE *)alloclptr(lines)) {
	    bmovl (BEp->List+BSline,list,lines);
	    bmovl (ep->List+ep->Line, ep->List+ep->Line+lines,
		    ep->Lines-ep->Line);

	    for (i = 0; i < lines; ++i) {
		str = allocline(strlen((char *)list[i])+1);

		if (!str) {
		    nomemory ();
		    FreeMem(list, lines * sizeof(LINE));
		    freelist(ep->List + ep->Line, i);
		    bmovl(ep->List+ep->Line+lines, ep->List+ep->Line, ep->Lines-ep->Line);
		    return;
		}

		strcpy((char *)str, (char *)list[i]);
		ep->List[ep->Line+i] = str;
	    }

	    FreeMem(list, lines * sizeof(LINE));
	}
    }

    if (ep == BEp && ep->Line <= BSline) {
	BSline += lines;
	BEline += lines;
    }

    ep->Modified = 1;
    ep->Lines += lines;

    text_load ();

    if (!text_sync ()) {
	scroll_display (0, -lines, 0, ep->Line, MAXLINELEN, ep->Topline+Rows);
    }
} /* do_bcopy() */


void do_bmove()
{
    long lines;
    LINE *list;
    ED *ep = Ep;

    /* TODO */
    if (BlockType != BT_LINE)

    text_sync ();

    if (!block_ok ())
	return;

    if (BEp == ep && ep->Line >= BSline && ep->Line <= BEline) {
	title("Cannot Move into self");
	return;
    }

    lines = BEline - BSline + 1;

    if (!(list = (LINE *)alloclptr (lines))) {
	nomemory ();
	return;
    }

    BEp->Modified = ep->Modified = 1;
    bmovl (BEp->List + BSline, list, lines);

    if (ep == BEp) {
	if (ep->Line > BSline) {
	    bmovl (ep->List+BEline+1, ep->List+BSline, ep->Line-BEline-1);
	    bmovl (list, ep->List + ep->Line - lines, lines);
	} else {
	    bmovl (ep->List+ep->Line, ep->List+ep->Line+lines, BSline-ep->Line);
	    bmovl (list, ep->List + ep->Line, lines);
	}
    } else {
	WIN *savewin = ep->Win;

	if (extend(ep, lines)) {
	    bmovl (BEp->List+BEline+1, BEp->List+BSline, BEp->Lines-BEline-1);
	    bmovl (ep->List+ep->Line, ep->List+ep->Line+lines, ep->Lines-ep->Line);
	    bmovl (list, ep->List+ep->Line, lines);

	    ep->Lines += lines;
	    BEp->Lines -= lines;

	    if (BEp->Line >= BSline && BEp->Line <= BEline)
		BEp->Line = BSline - 1;

	    if (BEp->Line > BEline)
		BEp->Line -= lines;

	    if (BEp->Line < 0)
		BEp->Line = 0;

	    BlockType = BT_NONE;

	    if (BEp->Lines == 0) {
		LINE ptr = allocline (1);
		BEp->List[0] = ptr;
		++BEp->Lines;
	    }

	    text_load ();
	    text_switch (BEp->Win);

	    BEp = NULL;

	    if (!Ep->iconmode) {
		if (!text_sync ())
		    text_redisplay ();
	    }

	    text_switch (savewin);
	    ep = Ep;
	}
    }

    BlockType = BT_NONE;

    FreeMem(list, lines * sizeof(LINE));
    ep->Modified = 1;

    text_load ();

    if (!text_sync ())
	text_redisplay ();
} /* do_bmove() */


/*
 *  BSOURCE
 *
 *  note that since the start and end lines are loaded immediately and the
 *  block unblock'd before execution starts, you can theoretically have
 *  another BSOURCE as part of this BSOURCE (but be carefull!).
 */

void do_bsource (void_IN  char buf[MAXLINELEN];
    int i, sl, se;

    /* TODO */
    if (BlockType != BT_LINE)
	return;

    if (block_ok ()) {
	sl = BSline;
	se = BEline + 1;

	for (i = sl; BEp && i < se && i < BEp->Lines; ++i) {
	    text_sync ();        /* make sure we are using latest text */
	    strcpy (buf, (char *)BEp->List[i]);
	    if (*buf != '#' && do_command (buf) == 0)
		break;
	}
	text_redrawblock (0);
    }
} /* do_bsource */


/******************************************************************************
*****  ENDE block.c
******************************************************************************/
rn;

    